"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const abstract_leveldown_1 = require("abstract-leveldown");
const level_mount_iterator_1 = require("./level-mount-iterator");
class LevelDOWNMount extends abstract_leveldown_1.AbstractLevelDOWN {
    constructor(root, location) {
        super(location);
        this._mounts = {};
        // Mount the root store to an empty prefix
        this.mount('', root);
    }
    // Mount a given prefix to an AbstractLevelDOWN-compatible store
    mount(mount_loc, mount) {
        this._mounts[mount_loc] = mount;
    }
    // Resolve the mount and internal path that a given key falls under
    _resolve_mount(key) {
        // TODO: improve mount search complexity?
        // TODO: handle keys that aren't strings?
        // Coerce key to string
        const key_str = String(key);
        // Sort mount locations by length (largest first)
        const mounts = Object.keys(this._mounts).sort((a, b) => b.length - a.length);
        // Select the first mount which matches the prefix of the key
        //  (the root mount with prefix '' will get matched as a fallback)
        for (const mount_loc of mounts) {
            if (key_str.slice(0, mount_loc.length) === mount_loc) {
                return {
                    mount: mount_loc,
                    path: key_str.slice(mount_loc.length),
                };
            }
        }
        throw new Error('An impossible error occured');
    }
    _open(options, callback) {
        process.nextTick(callback);
    }
    _get(key, options, callback) {
        // Forward to underlying store
        const { mount, path } = this._resolve_mount(key);
        this._mounts[mount].get(path, options, callback);
    }
    _put(key, val, options, callback) {
        // Forward to underlying store
        const { mount, path } = this._resolve_mount(key);
        this._mounts[mount].put(path, val, options, callback);
    }
    _del(key, options, callback) {
        // Forward to underlying store
        const { mount, path } = this._resolve_mount(key);
        this._mounts[mount].del(path, options, callback);
    }
    _batch(array, options, cb) {
        // Group batch operations by resolved mount
        const ops = array.reduce((grouped_ops, op) => {
            const { mount, path } = this._resolve_mount(op.key);
            if (grouped_ops[mount] === undefined)
                grouped_ops[mount] = [];
            grouped_ops[mount].push(Object.assign({}, op, { key: path }));
            return grouped_ops;
        }, {});
        // Forward batch options to underlying stores
        Promise.all(Object.keys(ops).map((mount) => new Promise((resolve, reject) => {
            this._mounts[mount].batch(ops[mount], options, (err) => {
                if (err)
                    reject(err);
                else
                    resolve();
            });
        }))).then(() => process.nextTick(cb)).catch((err) => process.nextTick(cb, err));
    }
    _iterator(options) {
        return new level_mount_iterator_1.LevelDOWNMountIterator(this, options);
    }
}
exports.LevelDOWNMount = LevelDOWNMount;
//# sourceMappingURL=level-mount.js.map